// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:matrix_sdk/src/event/room/message_event.dart';
import 'package:matrix_sdk/src/event/room/state/avatar_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/member_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/name_change_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/util/json.dart';
import 'package:meta/meta.dart';

abstract class Event extends Identifiable<EventId> implements JsonWritable {
  @protected
  String get matrixType;

  @override
  final EventId id;

  final SentState sentState = SentState.sent;

  final UnsignedData unsigned;

  EventContent get content;

  Event(this.id, this.unsigned);

  /// Create event from the given [json].
  ///
  /// If [user] is specified, will use that instance as
  /// user state info.
  ///
  /// Else, if [users] is specified, will look in that map
  /// and add possible user state info to the event.
  ///
  /// Otherwise it will create a new [User] instance with just
  /// the id.
  static Future<Event> fromJson(
      Room room,
      Map<String, dynamic> json,
      {User user}) async {
    var type = json['type'];

    // These might be null
    final id = EventId(json['event_id']);

    final senderId = UserId(json['sender']);
    var sender;

    if (user != null) {
      sender = user;
    } else {
      sender = await room.members[senderId] ?? User(id: senderId);
    }

    final stateKey = json['state_key'];

    final originServerTs = json['origin_server_ts'];
    var time;
    if (originServerTs != null) {
      time = DateTime.fromMillisecondsSinceEpoch(originServerTs);
    }

    var unsigned = json['unsigned'];
    if (unsigned != null) {
      final transactionId = json['unsigned']['transaction_id'];

      if (transactionId != null) {
        unsigned = UnsignedData(transactionId: transactionId);
      } else {
        unsigned = null;
      }
    }

    switch (type) {
      case NameChangeEvent.type:
        return NameChangeEvent.fromJson(id, room, sender, time, unsigned, json);
      case AvatarChangeEvent.type:
        return AvatarChangeEvent.fromJson(id, room, sender, time, unsigned, json);
      case MemberChangeEvent.type:
        return MemberChangeEvent.fromJson(id, room, sender, time, unsigned, stateKey, json);
      case MessageEvent.type:
        return MessageEvent.fromJson(id, room, sender, time, unsigned, json);
      default:
        return null;
    }
  }

  // TODO: Pass all room event args as single object
  factory Event.fromContent(EventContent content,
    {@required EventId id,
     @required Room room,
     @required DateTime time,
     @required User sender,
     @required UnsignedData unsigned,
      SentState sentState = SentState.sent}) {
    if (content is TextMessage) {
      return TextMessageEvent(
        content,
        id,
        room,
        sender,
        time,
        unsigned,
        sentState: sentState
      );
    } else {
      return null;
    }
  }

  @override
  Map<String, dynamic> toJson() => {
    'event_id': id.toJson(),
    'type': matrixType,
    'content': content.toJson()
  };

  @override
  String toString() => toJson().toString();
}

enum SentState {
  waiting,
  sent
}

abstract class EventContent {

  String get matrixType;

  Map<String, dynamic> toJson() { return {}; }
}

class UnsignedData {
  final String transactionId;

  UnsignedData({@required this.transactionId});
}