// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:meta/meta.dart';

const String _roomMemberType = 'm.room.member';

abstract class MemberChangeEvent extends StateEvent {
  static const String type = _roomMemberType;

  @override
  final String matrixType = type;

  @override
  final MemberChange content;

  MemberChangeEvent(
      this.content, EventId id,
      Room room, User sender, DateTime time, UnsignedData unsigned)
      : super(id, room, sender, time, unsigned, stateKey: content.subjectId.toString());

  factory MemberChangeEvent.fromJson(EventId id, Room room, User sender,
      DateTime time, UnsignedData unsigned, String stateKey, Map<String, dynamic> json) {
    final membershipType = json['content']['membership'];
    final subject = UserId(stateKey);
    final displayName = json['content']['displayname']
                     ?? sender.name
                     ?? json['prev_content']['displayname'];

    switch (membershipType) {
      case JoinEvent.type:
        return JoinEvent(subject, displayName, id, room, sender, time, unsigned);
      case LeaveEvent.type:
        return LeaveEvent(subject, displayName, id, room, sender, time, unsigned);
      case BanEvent.type:
        return BanEvent(subject, displayName, id, room, sender, time, unsigned);
      case InviteEvent.type:
        return InviteEvent(subject, displayName, id, room, sender, time, unsigned);
      case KnockEvent.type:
      default:
        return KnockEvent(subject, displayName, id, room, sender, time, unsigned);
    }
  }
}

class MemberChange extends StateEventContent {
  @override
  final String matrixType = _roomMemberType;

  @override
  final String stateKey;

  /// The [UserId] of who has joined, left, was banned, etc.
  final UserId subjectId;

  /// Display name of the [subjectId]. Might be null.
  final String displayName;

  final String membershipType;

  MemberChange(
    {@required this.subjectId, @required this.membershipType, this.displayName})
    : stateKey = subjectId.toString();

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'membership': membershipType,
      'displayname': displayName,
    });
}

class JoinEvent extends MemberChangeEvent {
  static const String type = 'join';

  JoinEvent(
      UserId subjectId, String displayName, EventId id,
      Room room, User sender, DateTime time, UnsignedData unsigned)
      : super(MemberChange(
        subjectId: subjectId,
        membershipType: type,
        displayName: displayName
      ), id, room, sender, time, unsigned);
}

class LeaveEvent extends MemberChangeEvent {
  static const String type = 'leave';

  LeaveEvent(UserId subjectId, String displayName, EventId id,
             Room room, User sender, DateTime time, UnsignedData unsigned)
      : super(MemberChange(
      subjectId: subjectId,
      membershipType: type,
      displayName: displayName
  ), id, room, sender, time, unsigned);
}

class BanEvent extends MemberChangeEvent {
  static const String type = 'ban';

  BanEvent(UserId subjectId, String displayName, EventId id,
      Room room, User sender, DateTime time, UnsignedData unsigned)
      : super(MemberChange(
      subjectId: subjectId,
      membershipType: type,
      displayName: displayName
  ), id, room, sender, time, unsigned);
}

// TODO: Add RoomState
class InviteEvent extends MemberChangeEvent {
  static const String type = 'invite';

  InviteEvent(UserId subjectId, String displayName, EventId id,
              Room room, User sender, DateTime time, UnsignedData unsigned)
      : super(MemberChange(
      subjectId: subjectId,
      membershipType: type,
      displayName: displayName
  ), id, room, sender, time, unsigned);
}

class KnockEvent extends MemberChangeEvent {
  static const String type = 'knock';

  KnockEvent(UserId subjectId, String displayName, EventId id,
             Room room, User sender, DateTime time, UnsignedData unsigned)
      : super(MemberChange(
      subjectId: subjectId,
      membershipType: type,
      displayName: displayName
  ), id, room, sender, time, unsigned);
}